% h1 example: Peterson's mutual exclusion algorithm, process 2.
%   Copyright (C) 2005 Jean Goubault-Larrecq and LSV, CNRS UMR 8643 & ENS Cachan.
%
%   This file is part of h1.
%
%   h1 is free software; you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation; either version 2, or (at your option)
%   any later version.
%
%   h1 is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with h1; see the file COPYING.  If not, write to
%   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

% Process 2 starts in state ncrit, with C2=true.
input_clause(init, axiom,
		   [--one_or_two(Tour),
		    ++reach(init, S1, ncrit, C1, true, Tour)]).

input_clause(one_or_two_one, axiom,
	     [++one_or_two(one)]).

input_clause(one_or_two_two, axiom,
	     [++one_or_two(two)]).

input_clause(bool_true, axiom,
	     [++bool(true)]).

input_clause(bool_false, axiom,
	     [++bool(false)]).

% At each state S2, process 2 may do nothing, and let
% S1, C1 and Tour be modified arbitrarily (i.e., process 1
% executes, and does whatever it is meant to do).
input_clause(idle2, axiom,
	     [--reach(X, S1,S2,C1,C2,Tour),
	      --one_or_two(Tour'),
	      --bool(C'1),
	      ++reach(s1(X), S'1,S2,C'1,C2,Tour')]).

% ncrit ---C2:=true---> attempt
input_clause(ncrit_set_c2, axiom,
	     [--reach(X, S1, ncrit, C1, C2, Tour),
	      ++reach(s2(X), S1, attempt, C1, true, Tour)]).

% attempt ---Tour:=1---> wait
input_clause(attempt_set_tour, axiom,
	     [--reach(X, S1, attempt, C1, C2, Tour),
	      ++reach(s2(X), S1, wait, C1, C2, one)]).

% attempt: while (C1==true && Tour==1);
% This clause is, of course, useless.
input_clause(wait, axiom,
	     [--reach(X, S1, wait, true, C2, one),
	      ++reach(s2(X), S1, wait, true, C2, one)]).

% wait ---if C1==false---> crit
input_clause(acquire, axiom,
	     [--reach(X, S1, wait, false, C2, Tour),
	      ++reach(s2(X), S1, crit, false, C2, Tour)]).

% wait ---if Tour==2---> crit
input_clause(acquire, axiom,
	     [--reach(X, S1, wait, C1, C2, two),
	      ++reach(s2(X), S1, crit, C1, C2, two)]).

% crit ---C2:=false---> ncrit
input_clause(release, axiom,
	     [--reach(X, S1, crit, C1, C2, Tour),
	      ++reach(s2(X), S1, ncrit, C1, false, Tour)]).
